/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Topological sort. 
 */

define top_test(a_list) = exists Rdy_list,O_list : {
    fixed_list(Rdy_list,|a_list|) and
    fixed_list(O_list,|a_list|) and
    forall i<|a_list| : {
	Rdy_list[i] = false and
	O_list[i] = false and
	sort_process(i,a_list,Rdy_list)
    } and
    O_list gets Rdy_list and
    halt(forall i<|Rdy_list| : Rdy_list[i]) and
    always output (difference(Rdy_list,O_list))
}.

define schedule_test(constraints) = exists Schedule,Phase : {
    fixed_list(Schedule,|constraints|) and
    make_schedule(|constraints|,constraints,Schedule,Phase) and
    always output(Schedule)
}.

define make_schedule(n,c,Schedule,Phase) = {
    forall i<n : {
	waiting(i,c,Schedule);
	active(i,Phase,Schedule);
	done(i,Schedule)
    } and
    Phase = 0 and
    Phase gets Phase+1 and
    halt (forall i<n : |Schedule[i]|=1)
}.

define waiting(i,c,Schedule) = {
    Schedule[i] = [] and
    stable (Schedule[i]) and
    halt(forall j<|c[i]| : |Schedule[c[i][j]]|=1)
}.

define active(i,Phase,Schedule) = {
    Schedule[i] <- [Phase] and
    keep_stable(Schedule[i]) and
    len(2)
}.

define done(i,Schedule) = {
    stable (Schedule[i])
}.

define keep_stable(X) = exists x : {x=X and keep(X=x)}.

define sort_process(i,a_list,Rdy_list) = {
    stable(Rdy_list[i]) and halt(forall j<|a_list[i]|:Rdy_list[a_list[i][j]]);
    skip and Rdy_list[i] := true; /* Don't use semicolon?? */
    stable(Rdy_list[i])
}.

define difference(L1,L2) = diff(|L1|-1,L1,L2).

define diff(i,L1,L2) = (if i<0 then []
		else diff(i-1,L1,L2) + (if L1[i]=L2[i] then [] else [i])).

define top_test1 = [[1],[2],[]].

define top_test2 = [[2],[7],[6],[5],[],[],[4,3],[5],[1,4]].

load "lists".
