/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * test of fuzzy operators using plot
 */
 
load "fuzzy_logic".
load "../library/conversion".
load "../library/tcl".

/* tcl fuzzy_3d */

define mf_X = trapezoid_mf($1$, $5$, $7$, $8$).

define mf_Y = trapezoid_mf($1$, $10$, $15$, $20$).

define plot_matrix(L1,L2) = {
 if L1 = [] then []
   else {
     if L2 = [] then []
       else {
         [ [ [L1[i], L2[j]]  : j<|L2| ] : i<|L1| ]
       }
   }
}.

define map_plot_matrix(op, L1, L2) = {
 if L1 = [] then []
   else {
     if L2 = [] then []
       else {
         [ [ op(L1[i], L2[j]) : j<|L2| ] : i<|L1| ]
       }
   }
}.

/* run */ define test_plot_matrix_1() = {
  exists L1,L2,X : {
    L1 = [ i : i < 50] and L2 = [ i : i < 25] and
    X = map_plot_matrix(plus, L1, L2) and
    output X and empty
  }
}.


/* run */ define test_1() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$, $0.25$, $10$) and
      Y = lrange_fl($0$, $0.75$, $30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_min, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_2() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_product, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
  }
}.

/* run */ define test_3() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_lukasiewicz, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_4() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_drastic, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_5() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_nilpotent_min, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_6() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_hamacher, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_7() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_einstein, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_8() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(and_min, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_9() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_sum, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_10() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_lukasiewicz, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_11() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_drastic, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_12() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_nilpotent_max, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_13() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_hamacher, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_14() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(or_einstein, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.


/* run */ define test_16() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(b_difference, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_17() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(b_sum, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_18() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_larsen, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_19() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_mamdani, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_20() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_min, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_21() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_goguen, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_22() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_lukasiewicz, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_23() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_nilpotent_min, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_24() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_standard_strict, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.

/* run */ define test_25() = {
  exists X,Y,Mu_X,Mu_Y,Z : {
    {
      X = lrange_fl($0$,$0.25$,$10$) and
      Y = lrange_fl($0$,$0.75$,$30$) and
      Mu_X = map(mf_X, X) and
      Mu_Y = map(mf_Y, Y) and
      Z = map_plot_matrix(imp_dkl, Mu_X, Mu_Y) and
      tcl("init", [$0$, $10$, $0$, $30$] ) and
      (forall e in Z :  tcl("add_to_data",e) ) and
      empty
    };
    {tcl("plot3d",[0]) and empty}
 }
}.













































